// This will be overwritten by ./configure

#define GEOGRAPHICLIB_VERSION_STRING "1.32"
#define GEOGRAPHICLIB_VERSION_MAJOR 1
#define GEOGRAPHICLIB_VERSION_MINOR 32
#define GEOGRAPHICLIB_VERSION_PATCH 0

// Undefine HAVE_LONG_DOUBLE if this type is unknown to the compiler
#define HAVE_LONG_DOUBLE 1

// Define WORDS_BIGENDIAN to be 1 if your machine is big endian
/* #undef WORDS_BIGENDIAN */
